@@ -3,6 +3,7 @@ use std::collections::{hash_map, HashMap};
 use std::fmt;
 use std::hash::{Hash, Hasher};
 use std::marker::Unpin;
+use std::net::SocketAddr;
 use std::sync::atomic;
 use std::sync::Arc;
 use std::sync::RwLock;
@@ -10,11 +11,14 @@ use std::time::Duration;
 use submap::{BroadcastMap, SubMap};
 use tokio::io::{AsyncReadExt, AsyncWriteExt, BufReader, BufWriter};
 use tokio::net::{TcpListener, TcpStream, UnixListener, UnixStream};
+#[cfg(feature = "broker-api")]
+use tokio::sync::Mutex;
 use tokio::task::JoinHandle;
 use tokio::time;
 
 use crate::{Error, ErrorKind, GREETINGS, PROTOCOL_VERSION};
 
+use crate::ERR_DATA;
 use crate::ERR_NOT_SUPPORTED;
 use crate::RESPONSE_OK;
 
@@ -25,9 +29,15 @@ use crate::client::AsyncClient;
 use crate::{EventChannel, OpConfirm};
 use crate::{Frame, FrameData, FrameKind, FrameOp, QoS};
 
+#[cfg(feature = "broker-api")]
+use crate::rpc::{Rpc, RpcClient, RpcError, RpcEvent, RpcHandlers, RpcResult};
+
 use async_trait::async_trait;
 
-const DEFAULT_QUEUE_SIZE: usize = 8192;
+pub const DEFAULT_QUEUE_SIZE: usize = 8192;
+
+pub const BROKER_INFO_TOPIC: &str = ".broker/info";
+pub const BROKER_WARN_TOPIC: &str = ".broker/warn";
 
 macro_rules! pretty_error {
     ($name: expr, $err:expr) => {
@@ -265,9 +275,45 @@ impl Drop for Client {
     }
 }
 
+#[derive(Copy, Clone, Debug, Eq, PartialEq)]
+enum ElbusClientType {
+    Internal,
+    LocalIpc,
+    Tcp,
+}
+
+impl ElbusClientType {
+    fn as_str(&self) -> &str {
+        match self {
+            ElbusClientType::Internal => "internal",
+            ElbusClientType::LocalIpc => "local_ipc",
+            ElbusClientType::Tcp => "tcp",
+        }
+    }
+}
+
+impl fmt::Display for ElbusClientType {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "{}", self.as_str())
+    }
+}
+
+#[cfg(feature = "broker-api")]
+impl serde::Serialize for ElbusClientType {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: serde::Serializer,
+    {
+        serializer.serialize_str(self.as_str())
+    }
+}
+
 #[derive(Debug)]
 struct ElbusClient {
     name: String,
+    tp: ElbusClientType,
+    source: Option<String>,
+    port: Option<String>,
     tx: async_channel::Sender<Frame>,
 }
 
@@ -278,11 +324,20 @@ impl fmt::Display for ElbusClient {
 }
 
 impl ElbusClient {
-    pub fn new(name: &str, queue_size: usize) -> (Self, EventChannel) {
+    pub fn new(
+        name: &str,
+        queue_size: usize,
+        tp: ElbusClientType,
+        source: Option<String>,
+        port: Option<String>,
+    ) -> (Self, EventChannel) {
         let (tx, rx) = async_channel::bounded(queue_size);
         (
             Self {
                 name: name.to_owned(),
+                tp,
+                source,
+                port,
                 tx,
             },
             rx,
@@ -336,13 +391,15 @@ impl BrokerDb {
             {
                 let mut sdb = self.subscriptions.write().unwrap();
                 sdb.register_client(&client);
-                sdb.subscribe("!bus", &client);
-                sdb.subscribe("!bus/#", &client);
+                sdb.subscribe(BROKER_WARN_TOPIC, &client);
             }
             x.insert(client);
             Ok(())
         } else {
-            Err(Error::busy("client already registred"))
+            Err(Error::busy(format!(
+                "the client is already registred: {}",
+                client.name
+            )))
         }
     }
     fn unregister_client(&self, client: &Arc<ElbusClient>) {
@@ -362,15 +419,66 @@ pub struct Broker {
     db: Arc<BrokerDb>,
     services: Vec<JoinHandle<()>>,
     queue_size: usize,
+    #[cfg(feature = "broker-api")]
+    rpc_client: Option<Arc<Mutex<RpcClient>>>,
+}
+
+#[cfg(feature = "broker-api")]
+struct BrokerRpcHandlers {
+    db: Arc<BrokerDb>,
+}
+
+#[cfg(feature = "broker-api")]
+#[async_trait]
+impl RpcHandlers for BrokerRpcHandlers {
+    async fn handle_call(&self, event: RpcEvent) -> RpcResult {
+        match event.parse_method()? {
+            "list_clients" => {
+                #[derive(serde::Serialize, Eq, PartialEq)]
+                struct ClientInfo<'a> {
+                    name: &'a str,
+                    tp: ElbusClientType,
+                    source: Option<&'a str>,
+                    port: Option<&'a str>,
+                }
+                impl<'a> Ord for ClientInfo<'a> {
+                    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
+                        self.name.cmp(other.name)
+                    }
+                }
+                impl<'a> PartialOrd for ClientInfo<'a> {
+                    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
+                        Some(self.cmp(other))
+                    }
+                }
+                #[derive(serde::Serialize)]
+                struct Clients<'a> {
+                    clients: Vec<ClientInfo<'a>>,
+                }
+                let db = self.db.clients.read().unwrap();
+                let mut clients: Vec<ClientInfo> = db
+                    .values()
+                    .into_iter()
+                    .map(|v| ClientInfo {
+                        name: &v.name,
+                        tp: v.tp,
+                        source: v.source.as_deref(),
+                        port: v.port.as_deref(),
+                    })
+                    .collect();
+                clients.sort();
+                Ok(Some(rmp_serde::to_vec_named(&Clients { clients })?))
+            }
+            _ => Err(RpcError::method()),
+        }
+    }
+    async fn handle_notification(&self, _event: RpcEvent) {}
+    async fn handle_frame(&self, _frame: Frame) {}
 }
 
 impl Default for Broker {
     fn default() -> Self {
-        Self {
-            db: <_>::default(),
-            services: <_>::default(),
-            queue_size: DEFAULT_QUEUE_SIZE,
-        }
+        Self::new()
     }
 }
 
@@ -385,9 +493,19 @@ fn prepare_tcp_stream(stream: &TcpStream) -> Result<(), Error> {
     stream.set_nodelay(true).map_err(Into::into)
 }
 
+#[allow(clippy::unnecessary_wraps)]
+fn prepare_tcp_source(addr: SocketAddr) -> Option<String> {
+    Some(addr.to_string())
+}
+
+#[allow(clippy::unnecessary_wraps)]
+fn prepare_unix_source(_addr: tokio::net::unix::SocketAddr) -> Option<String> {
+    None
+}
+
 macro_rules! spawn_server {
-    ($self: expr, $path: expr, $listener: expr, $buf_size: expr, $timeout: expr,
-     $prepare: ident) => {{
+    ($self: expr, $path: expr, $listener: expr, $buf_size: expr, $timeout: expr, $tp: expr,
+     $prepare: ident, $prepare_source: ident) => {{
         let socket_path = $path.to_owned();
         let db = $self.db.clone();
         let queue_size = $self.queue_size;
@@ -409,9 +527,20 @@ macro_rules! spawn_server {
                         let writer = BufWriter::with_capacity($buf_size, writer);
                         let cdb = db.clone();
                         let name = socket_path.clone();
+                        let client_source = $prepare_source(addr);
+                        let client_path = socket_path.clone();
                         tokio::spawn(async move {
-                            if let Err(e) =
-                                Self::handle_peer(cdb, reader, writer, $timeout, queue_size).await
+                            if let Err(e) = Self::handle_peer(PeerHandlerParams {
+                                db: cdb,
+                                reader,
+                                writer,
+                                timeout: $timeout,
+                                queue_size,
+                                tp: $tp,
+                                source: client_source,
+                                source_port: Some(client_path),
+                            })
+                            .await
                             {
                                 pretty_error!(name, e);
                             }
@@ -425,12 +554,53 @@ macro_rules! spawn_server {
     }};
 }
 
+struct PeerHandlerParams<R, W>
+where
+    R: AsyncReadExt + Unpin,
+    W: AsyncWriteExt + Unpin + Send + 'static,
+{
+    db: Arc<BrokerDb>,
+    reader: R,
+    writer: W,
+    timeout: Duration,
+    queue_size: usize,
+    tp: ElbusClientType,
+    source: Option<String>,
+    source_port: Option<String>,
+}
+
 impl Broker {
     pub fn new() -> Self {
-        Self::default()
+        let broker_db: Arc<BrokerDb> = <_>::default();
+        let mut broker = Self {
+            #[cfg(feature = "broker-api")]
+            db: broker_db.clone(),
+            #[cfg(not(feature = "broker-api"))]
+            db: broker_db,
+            services: <_>::default(),
+            queue_size: 0,
+            #[cfg(feature = "broker-api")]
+            rpc_client: None,
+        };
+        // avoid warning if rpc feature is not set
+        broker.queue_size = DEFAULT_QUEUE_SIZE;
+        #[cfg(feature = "broker-api")]
+        {
+            let client = broker
+                .register_client(".broker")
+                .expect("can not register broker RPC");
+            let handlers = BrokerRpcHandlers { db: broker_db };
+            let rpc_client = RpcClient::new(client, handlers);
+            broker.rpc_client.replace(Arc::new(Mutex::new(rpc_client)));
+        }
+        broker
     }
-    pub async fn register_client(&self, name: &str) -> Result<Client, Error> {
-        let (c, rx) = ElbusClient::new(name, self.queue_size);
+    pub fn set_queue_size(&mut self, queue_size: usize) {
+        self.queue_size = queue_size;
+    }
+    pub fn register_client(&self, name: &str) -> Result<Client, Error> {
+        let (c, rx) =
+            ElbusClient::new(name, self.queue_size, ElbusClientType::Internal, None, None);
         let client = Arc::new(c);
         self.db.register_client(client.clone())?;
         Ok(Client {
@@ -447,7 +617,16 @@ impl Broker {
     ) -> Result<(), Error> {
         let _r = tokio::fs::remove_file(path).await;
         let listener = UnixListener::bind(path)?;
-        spawn_server!(self, path, listener, buf_size, timeout, prepare_unix_stream);
+        spawn_server!(
+            self,
+            path,
+            listener,
+            buf_size,
+            timeout,
+            ElbusClientType::LocalIpc,
+            prepare_unix_stream,
+            prepare_unix_source
+        );
         Ok(())
     }
     pub async fn spawn_tcp_server(
@@ -457,11 +636,26 @@ impl Broker {
         timeout: Duration,
     ) -> Result<(), Error> {
         let listener = TcpListener::bind(path).await?;
-        spawn_server!(self, path, listener, buf_size, timeout, prepare_tcp_stream);
+        spawn_server!(
+            self,
+            path,
+            listener,
+            buf_size,
+            timeout,
+            ElbusClientType::Tcp,
+            prepare_tcp_stream,
+            prepare_tcp_source
+        );
         Ok(())
     }
     #[allow(clippy::items_after_statements)]
-    pub async fn spawn_fifo(&mut self, path: &str) -> Result<(), Error> {
+    #[cfg(feature = "broker-api")]
+    pub async fn spawn_fifo(&mut self, path: &str, buf_size: usize) -> Result<(), Error> {
+        let rpc_client = if let Some(ref c) = self.rpc_client {
+            c.clone()
+        } else {
+            return Err(Error::not_supported("broker RPC not initialized"));
+        };
         let _r = tokio::fs::remove_file(path).await;
         unix_named_pipe::create(path, Some(0o622))?;
         use std::os::unix::fs::PermissionsExt;
@@ -469,33 +663,138 @@ impl Broker {
         // chown fifo as it's usually created with 644
         tokio::fs::set_permissions(path, std::fs::Permissions::from_mode(0o622)).await?;
         let fd = unix_named_pipe::open_read(path)?;
-        //let socket_path = path.to_owned();
+        let socket_path = path.to_owned();
         let service = tokio::spawn(async move {
             let f = tokio::fs::File::from_std(fd);
-            let reader = BufReader::new(f);
-            let mut _lines = reader.lines();
-            //loop {
-            // TODO put rpc call
-            //while let Some(line) = lines.next_line().await.unwrap() {
-            //println!("{}", line);
-            //}
-            //}
+            let reader = BufReader::with_capacity(buf_size, f);
+            let mut lines = reader.lines();
+            let sleep_step = Duration::from_millis(100);
+            loop {
+                while let Some(line) = match lines.next_line().await {
+                    Ok(v) => v,
+                    Err(e) => {
+                        error!("{}: {}", socket_path, e);
+                        None
+                    }
+                } {
+                    if let Err(e) = Self::send_fifo_cmd(&rpc_client, line).await {
+                        error!("{}: {}", socket_path, e);
+                    }
+                }
+                tokio::time::sleep(sleep_step).await;
+            }
         });
         self.services.push(service);
         Ok(())
     }
+    #[cfg(feature = "broker-api")]
+    async fn send_fifo_cmd(rpc: &Arc<Mutex<RpcClient>>, line: String) -> Result<(), Error> {
+        let cmd = line.trim();
+        // topic
+        if let Some(s) = cmd.strip_prefix('=') {
+            let mut sp = s.split(' ');
+            let topic = sp
+                .next()
+                .ok_or_else(|| Error::data("topic not specified"))?;
+            let payload = sp
+                .next()
+                .ok_or_else(|| Error::data("payload not specified"))?;
+            rpc.lock()
+                .await
+                .client()
+                .lock()
+                .await
+                .publish(topic, payload.as_bytes().into(), QoS::No)
+                .await?;
+            Ok(())
+        } else {
+            let mut sp = line.split(' ');
+            let target = sp
+                .next()
+                .ok_or_else(|| Error::data("target not specified"))?;
+            let payload = sp
+                .next()
+                .ok_or_else(|| Error::data("payload not specified"))?;
+            // rpc notification
+            if let Some(s) = payload.strip_prefix('.') {
+                rpc.lock()
+                    .await
+                    .notify(target, s.as_bytes().into(), QoS::No)
+                    .await?;
+                Ok(())
+            } else if let Some(method) = payload.strip_prefix(':') {
+                use serde_json::value::Value;
+                let mut params: HashMap<&str, Value> = HashMap::new();
+                for pair in sp {
+                    if !pair.is_empty() {
+                        let mut psp = pair.split('=');
+                        let var = psp
+                            .next()
+                            .ok_or_else(|| Error::data("var name not specified"))?;
+                        let v = psp
+                            .next()
+                            .ok_or_else(|| Error::data("var value not specified"))?;
+                        let value = if v == "false" {
+                            Value::from(false)
+                        } else if v == "true" {
+                            Value::from(true)
+                        } else if let Ok(i) = v.parse::<i64>() {
+                            Value::from(i)
+                        } else if let Ok(f) = v.parse::<f64>() {
+                            Value::from(f)
+                        } else {
+                            Value::from(v)
+                        };
+                        params.insert(var, value);
+                    }
+                }
+                rpc.lock()
+                    .await
+                    .call0(
+                        target,
+                        method,
+                        rmp_serde::to_vec_named(&params)
+                            .map_err(Error::data)?
+                            .into(),
+                    )
+                    .await?;
+                Ok(())
+            } else {
+                // regular message
+                // broadcast
+                if target.contains(&['*', '?'][..]) {
+                    rpc.lock()
+                        .await
+                        .client()
+                        .lock()
+                        .await
+                        .send_broadcast(target, payload.as_bytes().into(), QoS::No)
+                        .await?;
+                    Ok(())
+                } else {
+                    rpc.lock()
+                        .await
+                        .client()
+                        .lock()
+                        .await
+                        .send(target, payload.as_bytes().into(), QoS::No)
+                        .await?;
+                    Ok(())
+                }
+            }
+        }
+    }
     #[allow(clippy::too_many_lines)]
-    async fn handle_peer<R, W>(
-        db: Arc<BrokerDb>,
-        mut reader: R,
-        mut writer: W,
-        timeout: Duration,
-        queue_size: usize,
-    ) -> Result<(), Error>
+    async fn handle_peer<R, W>(params: PeerHandlerParams<R, W>) -> Result<(), Error>
     where
         R: AsyncReadExt + Unpin,
         W: AsyncWriteExt + Unpin + Send + 'static,
     {
+        let timeout = params.timeout;
+        let mut reader = params.reader;
+        let mut writer = params.writer;
+        let queue_size = params.queue_size;
+        let db = params.db;
         macro_rules! write_and_flush {
             ($buf: expr) => {
                 time::timeout(timeout, writer.write_all($buf)).await??;
@@ -522,11 +821,18 @@ impl Broker {
         let mut buf = vec![0; len as usize];
         time::timeout(timeout, reader.read_exact(&mut buf)).await??;
         let client_name = std::str::from_utf8(&buf)?.to_owned();
-        if client_name.is_empty() || client_name == "!bus" {
+        if client_name.is_empty() || client_name.starts_with('.') {
+            write_and_flush!(&[ERR_DATA]);
             return Err(Error::data("Invalid client name"));
         }
         let (client, rx) = {
-            let (c, rx) = ElbusClient::new(&client_name, queue_size);
+            let (c, rx) = ElbusClient::new(
+                &client_name,
+                queue_size,
+                params.tp,
+                params.source,
+                params.source_port,
+            );
             let client = Arc::new(c);
             if let Err(e) = db.register_client(client.clone()) {
                 write_and_flush!(&[e.kind as u8]);
@@ -583,6 +889,9 @@ impl Broker {
                     if let Some(t) = topic.as_ref() {
                         extra_len += t.len() + 1;
                     }
+                    if let Some(header) = frame.header.as_ref() {
+                        extra_len += header.len();
+                    }
                     let mut buf = Vec::with_capacity(7 + extra_len);
                     buf.push(frame.kind as u8); // byte 0
                     let frame_len = extra_len + frame.buf.len() - frame.payload_pos + 1;
@@ -695,7 +1004,9 @@ impl Broker {
                     match op {
                         FrameOp::Message => {
                             if let Err(e) = send!(db, client, target, None, buf, payload_pos) {
-                                send_ack!(e.kind as u8);
+                                if qos == QoS::Processed {
+                                    send_ack!(e.kind as u8);
+                                }
                             } else if qos == QoS::Processed {
                                 send_ack!(RESPONSE_OK);
                             }
